﻿//============================================================================
// Name        : pkcs11test.cpp
// Author      : 
// Version     :
// Copyright   : Your copyright notice
// Description : Hello World in C++, Ansi-style
//============================================================================

#include <iostream>

#include "cryptoki.h"
#include "stdio.h"

#ifndef _WIN32
 #include <dlfcn.h>
#else
 #include <windows.h>
#endif

typedef unsigned long DWORD;

unsigned char ucPWDUser[] = { 0x31, 0x32, 0x33, 0x34, 0x35, 0x36 };
unsigned long dwLenPWDUser = 6;

void printBytes(CK_BYTE_PTR pBytes, DWORD dLenBytes) {

	for (DWORD d = 0; d < dLenBytes; d++) {
		printf("0x%X ",(pBytes[d]));
	}
}

CK_FUNCTION_LIST_PTR getP11(void** hDll, std::string sPath) {

	#ifndef _WIN32
	(*hDll) = dlopen(sPath.c_str(),RTLD_NOW);
	printf("\nLib Handle : 0x%lX",((unsigned long)(*hDll)));
	printf("\nError after dlsym : %s",dlerror());

	typedef void (*getFunctionList)(CK_FUNCTION_LIST_PTR_PTR  ppFunctionList);
	getFunctionList getFunctions = (getFunctionList) dlsym((*hDll), "C_GetFunctionList");
	printf("\nError after dlsym : %s",dlerror());

	CK_FUNCTION_LIST_PTR p11;
	getFunctions(&p11);
	#else
	CK_FUNCTION_LIST_PTR p11;
	CK_RV returnCode;

	//(*hDll) = LoadLibrary(sPath.c_str());
	(*hDll) = LoadLibrary("P11TCOS3NetKey.dll");
	CK_C_GetFunctionList C_GetFunctionListDyn = (CK_C_GetFunctionList) GetProcAddress(hDll, "C_GetFunctionList");
	C_GetFunctionListDyn(&p11);
	#endif
	return p11;
}

void freeDll(void* hDll) {
	
	#ifndef _WIN32
 	 dlclose(hDll);
	#else
	 FreeLibrary(hDll);
	#endif
}

int pkSignRecover(CK_FUNCTION_LIST_PTR p11, void* hDll) {

	CK_RV returnCode;
		
	   //PKCS#11 initialisieren
	   returnCode = p11->C_Initialize(NULL_PTR);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_Initialize");
			return -1;
	   } else {
	        printf("\nC_Initialize erfolgreich");
	   }

	   //Anzahl Slots holen
	   CK_ULONG ulSlotCount;
	   returnCode = p11->C_GetSlotList(CK_TRUE, NULL_PTR, &ulSlotCount);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetSlotList");
	        return -1;
	   } else {
	        printf("\nC_GetSlotList erfolgreich");
	   }

	   //Sind Slots da?
	   if (ulSlotCount == 0) {
	        printf("\nKeine Slots gefunden");
	        return -1;
	   } else {
	        printf("\n %ld Slots gefunden !",ulSlotCount);
	   }

	   //Slots holen
	   CK_SLOT_ID_PTR pSlotId = (CK_SLOT_ID_PTR) malloc(sizeof(CK_SLOT_ID) * ulSlotCount);
	   returnCode = p11->C_GetSlotList(CK_TRUE, pSlotId, &ulSlotCount);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_GetSlotList");
	        return -1;
	   } else {
	        printf("\nC_GetSlotList erfolgreich, verwende SlotID : %lX",pSlotId[0]);
	   }

	   //1. Slot verwenden !!
	   //Session �ffnen
	   CK_SESSION_HANDLE hSession;
	   returnCode = p11->C_OpenSession(pSlotId[0], CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession);
	   //War das OK?
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_OpenSession");
	        return -1;
	   } else {
			printf("\nC_OpenSession erfolgreich, Session Handle : %lX",hSession);
	   }

	   //Einloggen ohne PinPad
	   returnCode = p11->C_Login(hSession, CKU_USER, ucPWDUser, dwLenPWDUser);
	   //War das OK?
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_Login");
	        return -1;
	   } else {
			printf("\nC_Login erfolgreich, User: CKU_USER, PIN: '123456', Session: %lX",hSession);
	   }

	   //////////////////////////////////////
	   // Zertifikat suchen
	   // Template f�r X.509 Zertifikat
	   CK_OBJECT_CLASS cClass   = CKO_CERTIFICATE;
	   CK_CERTIFICATE_TYPE   certType  = CKC_X_509;
	   CK_ATTRIBUTE certTempl[]    = {
			{
				CKA_CLASS,
				(unsigned char *) &cClass,
				sizeof(cClass)
			},
			{
				CKA_CERTIFICATE_TYPE,
				(unsigned char *) &certType,
				sizeof(CK_CERTIFICATE_TYPE)
			}
	   };

	   returnCode = p11->C_FindObjectsInit(hSession,certTempl,2);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsInit");
	        return -1;
	   } else {
	        printf("\nC_FindObjectsInit erfolgreich, CKA_CLASS: CKO_CERTIFICATE, CKA_CERTIFICATE_TYPE: CKC_X_509");
	   }
	   CK_OBJECT_HANDLE pCertificateHandle;
	   CK_ULONG ulMaxCount;

	   // Erstes gefundene Zertifikat verwenden in pCertificateHandle
	   returnCode = p11->C_FindObjects(hSession,&pCertificateHandle,1,&ulMaxCount);
	   if ( (returnCode != CKR_OK) || (ulMaxCount==0)) {
	        printf("\nFehler bei C_FindObjects");
			return -1;
	   } else {
	        printf("\nC_FindObjects erfolgreich, CERTIFICATE Handle: %lX", pCertificateHandle);
	   }

	   // Suche abschliessen
	   returnCode = p11->C_FindObjectsFinal(hSession);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsFinal");
	        return -1;
	   } else {
	        printf("\nC_FindObjectsFinal erfolgreich");
	   }

	   // Zertifikatshandle nun in pCertificateHandle !!
	   // Anhand dessen CKA_ID finden
	   CK_BYTE pbCKA_ID[1024];
	   memset(pbCKA_ID,0x00,1024);
	   CK_ATTRIBUTE paKeyCKA_ID[]   = { { CKA_ID, NULL, 0 } };
	   p11->C_GetAttributeValue(hSession, pCertificateHandle, paKeyCKA_ID, 1);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetAttributeValue");
			return -1;
	   } else {
			printf("\nC_GetAttributeValue erfolgreich, CKA_ID L�nge: %ld", paKeyCKA_ID[0].ulValueLen);
	   }
	   DWORD dLenID = paKeyCKA_ID[0].ulValueLen;
	   CK_ATTRIBUTE pKeyCKA_ID[] = { { CKA_ID, pbCKA_ID, dLenID } };
	   p11->C_GetAttributeValue(hSession, pCertificateHandle, pKeyCKA_ID, 1);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetAttributeValue");
			return -1;
	   } else {
			printf("\nC_GetAttributeValue erfolgreich, CKA_ID :");
			printBytes(pbCKA_ID,dLenID);
			printf("\n");
	   }

	   //////////////////////////////////////
	   // PrivateKey anhand der CKA_ID des Zertifikats finden
	   cClass = CKO_PRIVATE_KEY;
	   CK_ATTRIBUTE keyTempl[]    = {
			{
				CKA_CLASS,
				(unsigned char *) &cClass,
				sizeof(cClass)
			},
			{
				CKA_ID,
			   (unsigned char *) &pbCKA_ID,
				dLenID
			}
	   };

	   returnCode = p11->C_FindObjectsInit(hSession,keyTempl,2);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsInit");
	        return -1;
	   } else {
			printf("\nC_FindObjectsInit erfolgreich, CKA_CLASS: CKO_PRIVATE_KEY CKA_ID: ");
			printBytes(pbCKA_ID,dLenID);
			printf("\n");
	   }

	   CK_OBJECT_HANDLE pObject;
	   returnCode = p11->C_FindObjects(hSession,&pObject,1,&ulMaxCount);
	   if ( (returnCode != CKR_OK) || (ulMaxCount==0))  {
			printf("\nFehler bei C_FindObjects");
			if (ulMaxCount==0)
				printf("\nKeinen privaten Schl�ssel zum Zertifikat gefunden !");
			return -1;
	   } else {
	        printf("\nC_FindObjects erfolgreich, Key Handle: %lX", pObject);
	   }

	   returnCode = p11->C_FindObjectsFinal(hSession);

	   //War das OK?
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsFinal");
	   }else {
	        printf("\nC_FindObjectsFinal erfolgreich");
	   }

	   //////////////////////////////////////
	   //Signatur beginnen
	   CK_MECHANISM mechanism;
	   CK_OBJECT_HANDLE hKey = pObject;
	   CK_BYTE data[1024], sig[256];
	   CK_ULONG lenSig=256;
	   mechanism.mechanism = CKM_SHA512_RSA_PKCS;
	   mechanism.pParameter = NULL_PTR;
	   mechanism.ulParameterLen = 0;

	   for(int i=0; i<1024; i++) data[i] = i % 256;

	   // Signaturvorgang initialisieren
	   returnCode = p11->C_SignRecoverInit(hSession, &mechanism, hKey);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei C_SignRecoverInit");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

	        return -1;
	     }   
	   
	   printf("\nC_SignRecoverInit erfolgreich");

	   //Signieren alternativ zu C_Sign
	   returnCode = p11->C_SignRecover(hSession, data, 1024, sig, &lenSig);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei C_SignRecover");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

			return -1;
		 }
	   printf("\nC_SignRecover erfolgreich !\n Signatur:\n");
	   printBytes(sig,lenSig);
	   printf("\n");

	   //Session schlie�en
	   returnCode = p11->C_CloseSession(hSession);

	   //War das OK?
	   if (returnCode != CKR_OK)
	     {
	        printf("\nFehler bei C_CloseSession");
	        free(pSlotId);
	        p11->C_Finalize(NULL_PTR);

	        return -1;
		 }

	   //Speicher freigeben
	   free(pSlotId);

	   //Sitzung beenden
	   returnCode = p11->C_Finalize(NULL_PTR);

	   if (returnCode != CKR_OK)
	     {
	        printf("\nFehler bei C_Finalize");

	        return -1;
	     }

	   //Warten
	   printf("\nTest erfolgreich !");

	   printf("\nC_SignRecoverInit erfolgreich");

	   //Signieren alternativ zu C_Sign
	   returnCode = p11->C_SignRecover(hSession, data, 1024, sig, &lenSig);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei C_SignRecover");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

			return -1;
		 }
	   printf("\nC_SignRecover erfolgreich !\n Signatur:\n");
	   printBytes(sig,lenSig);
	   printf("\n");

	   //Session schlie�en
	   returnCode = p11->C_CloseSession(hSession);

	   //War das OK?
	   if (returnCode != CKR_OK)
	     {
	        printf("\nFehler bei C_CloseSession");
	        free(pSlotId);
	        p11->C_Finalize(NULL_PTR);

	        return -1;
		 }

	   //Speicher freigeben
	   free(pSlotId);

	   //Sitzung beenden
	   returnCode = p11->C_Finalize(NULL_PTR);

	   if (returnCode != CKR_OK)
	     {
	        printf("\nFehler bei C_Finalize");

	        return -1;
	     }

	   //Warten
	   printf("\nTest erfolgreich !");
	   return 0;
	
}

int pkSign(CK_FUNCTION_LIST_PTR p11, void* hDll,bool bPinPad) {
	
		CK_RV returnCode;
		
	   //PKCS#11 initialisieren
	   returnCode = p11->C_Initialize(NULL_PTR);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_Initialize");
			return -1;
	   } else {
	        printf("\nC_Initialize erfolgreich");
	   }

	   //Anzahl Slots holen
	   CK_ULONG ulSlotCount;
	   returnCode = p11->C_GetSlotList(CK_TRUE, NULL_PTR, &ulSlotCount);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetSlotList");
	        return -1;
	   } else {
	        printf("\nC_GetSlotList erfolgreich");
	   }

	   //Sind Slots da?
	   if (ulSlotCount == 0) {
	        printf("\nKeine Slots gefunden");
	        return -1;
	   } else {
	        printf("\n %ld Slots gefunden !",ulSlotCount);
	   }

	   //Slots holen
	   CK_SLOT_ID_PTR pSlotId = (CK_SLOT_ID_PTR) malloc(sizeof(CK_SLOT_ID) * ulSlotCount);
	   returnCode = p11->C_GetSlotList(CK_TRUE, pSlotId, &ulSlotCount);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_GetSlotList");
	        return -1;
	   } else {
	        printf("\nC_GetSlotList erfolgreich, verwende SlotID : %lX",pSlotId[0]);
	   }

	   //1. Slot verwenden !!
	   //Session �ffnen
	   CK_SESSION_HANDLE hSession;
	   returnCode = p11->C_OpenSession(pSlotId[0], CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession);
	   //War das OK?
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_OpenSession");
	        return -1;
	   } else {
			printf("\nC_OpenSession erfolgreich, Session Handle : %lX",hSession);
	   }

	   //Einloggen ohne PinPad
	   if (bPinPad) {
		   returnCode = p11->C_Login(hSession, CKU_USER, NULL, 0);
	   } else {
		   returnCode = p11->C_Login(hSession, CKU_USER, ucPWDUser, dwLenPWDUser);
	   }
		
	   //War das OK?
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_Login");
	        return -1;
	   } else {
			printf("\nC_Login erfolgreich, User: CKU_USER, PIN: '123456', Session: %lX",hSession);
	   }

	   //////////////////////////////////////
	   // Zertifikat suchen
	   // Template f�r X.509 Zertifikat
	   CK_OBJECT_CLASS cClass   = CKO_CERTIFICATE;
	   CK_CERTIFICATE_TYPE   certType  = CKC_X_509;
	   CK_ATTRIBUTE certTempl[]    = {
			{
				CKA_CLASS,
				(unsigned char *) &cClass,
				sizeof(cClass)
			},
			{
				CKA_CERTIFICATE_TYPE,
				(unsigned char *) &certType,
				sizeof(CK_CERTIFICATE_TYPE)
			}
	   };

	   returnCode = p11->C_FindObjectsInit(hSession,certTempl,2);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsInit");
	        return -1;
	   } else {
	        printf("\nC_FindObjectsInit erfolgreich, CKA_CLASS: CKO_CERTIFICATE, CKA_CERTIFICATE_TYPE: CKC_X_509");
	   }
	   CK_OBJECT_HANDLE pCertificateHandle;
	   CK_ULONG ulMaxCount;

	   // Erstes gefundene Zertifikat verwenden in pCertificateHandle
	   returnCode = p11->C_FindObjects(hSession,&pCertificateHandle,1,&ulMaxCount);
	   if ( (returnCode != CKR_OK) || (ulMaxCount==0)) {
	        printf("\nFehler bei C_FindObjects");
			return -1;
	   } else {
	        printf("\nC_FindObjects erfolgreich, CERTIFICATE Handle: %lX", pCertificateHandle);
	   }

	   // Suche abschliessen
	   returnCode = p11->C_FindObjectsFinal(hSession);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsFinal");
	        return -1;
	   } else {
	        printf("\nC_FindObjectsFinal erfolgreich");
	   }

	   // Zertifikatshandle nun in pCertificateHandle !!
	   // Anhand dessen CKA_ID finden
	   CK_BYTE pbCKA_ID[1024];
	   memset(pbCKA_ID,0x00,1024);
	   CK_ATTRIBUTE paKeyCKA_ID[]   = { { CKA_ID, NULL, 0 } };
	   p11->C_GetAttributeValue(hSession, pCertificateHandle, paKeyCKA_ID, 1);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetAttributeValue");
			return -1;
	   } else {
			printf("\nC_GetAttributeValue erfolgreich, CKA_ID L�nge: %ld", paKeyCKA_ID[0].ulValueLen);
	   }
	   DWORD dLenID = paKeyCKA_ID[0].ulValueLen;
	   CK_ATTRIBUTE pKeyCKA_ID[] = { { CKA_ID, pbCKA_ID, dLenID } };
	   p11->C_GetAttributeValue(hSession, pCertificateHandle, pKeyCKA_ID, 1);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetAttributeValue");
			return -1;
	   } else {
			printf("\nC_GetAttributeValue erfolgreich, CKA_ID :");
			printBytes(pbCKA_ID,dLenID);
			printf("\n");
	   }

	   //////////////////////////////////////
	   // PrivateKey anhand der CKA_ID des Zertifikats finden
	   cClass = CKO_PRIVATE_KEY;
	   CK_ATTRIBUTE keyTempl[]    = {
			{
				CKA_CLASS,
				(unsigned char *) &cClass,
				sizeof(cClass)
			},
			{
				CKA_ID,
			   (unsigned char *) &pbCKA_ID,
				dLenID
			}
	   };

	   returnCode = p11->C_FindObjectsInit(hSession,keyTempl,2);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsInit");
	        return -1;
	   } else {
			printf("\nC_FindObjectsInit erfolgreich, CKA_CLASS: CKO_PRIVATE_KEY CKA_ID: ");
			printBytes(pbCKA_ID,dLenID);
			printf("\n");
	   }

	   CK_OBJECT_HANDLE pObject;
	   returnCode = p11->C_FindObjects(hSession,&pObject,1,&ulMaxCount);
	   if ( (returnCode != CKR_OK) || (ulMaxCount==0))  {
			printf("\nFehler bei C_FindObjects");
			if (ulMaxCount==0)
				printf("\nKeinen privaten Schl�ssel zum Zertifikat gefunden !");
			return -1;
	   } else {
	        printf("\nC_FindObjects erfolgreich, Key Handle: %lX", pObject);
	   }

	   returnCode = p11->C_FindObjectsFinal(hSession);

	   //War das OK?
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsFinal");
	   }else {
	        printf("\nC_FindObjectsFinal erfolgreich");
	   }

	   //////////////////////////////////////
	   //Signatur beginnen
	   CK_MECHANISM mechanism;
	   CK_OBJECT_HANDLE hKey = pObject;
	   CK_BYTE data[1024], sig[256];
	   CK_ULONG lenSig=256;
	   mechanism.mechanism = CKM_SHA512_RSA_PKCS;
	   mechanism.pParameter = NULL_PTR;
	   mechanism.ulParameterLen = 0;

	   for(int i=0; i<1024; i++) data[i] = i % 256;

	   // Signaturvorgang initialisieren
	   returnCode = p11->C_SignInit(hSession, &mechanism, hKey);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei C_SignInit");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

	        return -1;
	     }
	   printf("\nC_SignInit erfolgreich");

	   //Signieren alternativ zu C_Sign
	   returnCode = p11->C_SignUpdate(hSession, data, 1024);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei C_SignUpdate");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

			return -1;
		 }
	   printf("\nC_SignUpdate erfolgreich !\n");

	   returnCode = p11->C_SignFinal(hSession, sig, &lenSig);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei C_SignFinal");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

			return -1;
		 }

	   printf("\nC_SignFinal erfolgreich !\n Signatur:\n");
	   printBytes(sig,lenSig);
	   printf("\n");

	   //Session schlie�en
	   returnCode = p11->C_CloseSession(hSession);

	   //War das OK?
	   if (returnCode != CKR_OK)
	     {
	        printf("\nFehler bei C_CloseSession");
	        free(pSlotId);
	        p11->C_Finalize(NULL_PTR);

	        return -1;
		 }

	   //Speicher freigeben
	   free(pSlotId);

	   //Sitzung beenden
	   returnCode = p11->C_Finalize(NULL_PTR);

	   if (returnCode != CKR_OK)
	     {
	        printf("\nFehler bei C_Finalize");

	        return -1;
	     }

	   //Warten
	   printf("\nTest erfolgreich !");
	   return 0;
	
}

int pkEncrypt(CK_FUNCTION_LIST_PTR p11, void* hDll) {

		CK_RV returnCode;
		
	   //PKCS#11 initialisieren
	   returnCode = p11->C_Initialize(NULL_PTR);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_Initialize");
			return -1;
	   } else {
	        printf("\nC_Initialize erfolgreich");
	   }

	   //Anzahl Slots holen
	   CK_ULONG ulSlotCount;
	   returnCode = p11->C_GetSlotList(CK_TRUE, NULL_PTR, &ulSlotCount);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetSlotList");
	        return -1;
	   } else {
	        printf("\nC_GetSlotList erfolgreich");
	   }

	   //Sind Slots da?
	   if (ulSlotCount == 0) {
	        printf("\nKeine Slots gefunden");
	        return -1;
	   } else {
	        printf("\n %ld Slots gefunden !",ulSlotCount);
	   }

	   //Slots holen
	   CK_SLOT_ID_PTR pSlotId = (CK_SLOT_ID_PTR) malloc(sizeof(CK_SLOT_ID) * ulSlotCount);
	   returnCode = p11->C_GetSlotList(CK_TRUE, pSlotId, &ulSlotCount);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_GetSlotList");
	        return -1;
	   } else {
	        printf("\nC_GetSlotList erfolgreich, verwende SlotID : %lX",pSlotId[0]);
	   }

	   //1. Slot verwenden !!
	   //Session �ffnen
	   CK_SESSION_HANDLE hSession;
	   returnCode = p11->C_OpenSession(pSlotId[0], CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession);
	   //War das OK?
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_OpenSession");
	        return -1;
	   } else {
			printf("\nC_OpenSession erfolgreich, Session Handle : %lX",hSession);
	   }

	   //Einloggen ohne PinPad
	   returnCode = p11->C_Login(hSession, CKU_USER, ucPWDUser, dwLenPWDUser);
	   //War das OK?
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_Login");
	        return -1;
	   } else {
			printf("\nC_Login erfolgreich, User: CKU_USER, PIN: '123456', Session: %lX",hSession);
	   }

	   //////////////////////////////////////
	   // Zertifikat suchen
	   // Template f�r X.509 Zertifikat
	   CK_OBJECT_CLASS cClass   = CKO_PUBLIC_KEY;
	   CK_BBOOL cbEncrypt = CK_TRUE;
	   CK_ATTRIBUTE certTempl[]    = {
			{
				CKA_CLASS,
				&cClass,
				sizeof(cClass)
			},
			{
				CKA_ENCRYPT,
				 &cbEncrypt,
				sizeof(cbEncrypt)
			}
	   };

	   returnCode = p11->C_FindObjectsInit(hSession,certTempl,2);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsInit");
	        return -1;
	   } else {
			printf("\nC_FindObjectsInit erfolgreich, CKA_CLASS: CKO_PUBLIC_KEY, CKA_ENCRYPT: TRUE");
	   }
	   CK_OBJECT_HANDLE pPublic;
	   CK_ULONG ulMaxCount;

	   // Erstes gefundene Zertifikat verwenden in pCertificateHandle
	   returnCode = p11->C_FindObjects(hSession,&pPublic,1,&ulMaxCount);
	   if ( (returnCode != CKR_OK) || (ulMaxCount==0)) {
	        printf("\nFehler bei C_FindObjects");
			return -1;
	   } else {
			printf("\nC_FindObjects erfolgreich, Public Key Handle: %lX", pPublic);
	   }

	   // Suche abschliessen
	   returnCode = p11->C_FindObjectsFinal(hSession);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsFinal");
	        return -1;
	   } else {
	        printf("\nC_FindObjectsFinal erfolgreich");
	   }

	   // Zertifikatshandle nun in pCertificateHandle !!
	   // Anhand dessen CKA_ID finden
	   CK_BYTE pbCKA_ID[1024];
	   memset(pbCKA_ID,0x00,1024);
	   CK_ATTRIBUTE paKeyCKA_ID[]   = { { CKA_ID, NULL, 0 } };
	   p11->C_GetAttributeValue(hSession, pPublic, paKeyCKA_ID, 1);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetAttributeValue");
			return -1;
	   } else {
			printf("\nC_GetAttributeValue erfolgreich, CKA_ID L�nge: %ld", paKeyCKA_ID[0].ulValueLen);
	   }
	   DWORD dLenID = paKeyCKA_ID[0].ulValueLen;
	   CK_ATTRIBUTE pKeyCKA_ID[] = { { CKA_ID, pbCKA_ID, dLenID } };
	   p11->C_GetAttributeValue(hSession, pPublic, pKeyCKA_ID, 1);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetAttributeValue");
			return -1;
	   } else {
			printf("\nC_GetAttributeValue erfolgreich, CKA_ID :");
			printBytes(pbCKA_ID,dLenID);
			printf("\n");
	   }

	   //////////////////////////////////////
	   // PrivateKey anhand der CKA_ID des Zertifikats finden
	   cClass = CKO_PRIVATE_KEY;
	   CK_ATTRIBUTE keyTempl[]    = {
			{
				CKA_CLASS,
				(unsigned char *) &cClass,
				sizeof(cClass)
			},
			{
				CKA_ID,
			   (unsigned char *) &pbCKA_ID,
				dLenID
			}
	   };

	   returnCode = p11->C_FindObjectsInit(hSession,keyTempl,2);
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsInit");
	        return -1;
	   } else {
			printf("\nC_FindObjectsInit erfolgreich, CKA_CLASS: CKO_PRIVATE_KEY CKA_ID: ");
			printBytes(pbCKA_ID,dLenID);
			printf("\n");
	   }

	   CK_OBJECT_HANDLE pObject;
	   returnCode = p11->C_FindObjects(hSession,&pObject,1,&ulMaxCount);
	   if ( (returnCode != CKR_OK) || (ulMaxCount==0))  {
			printf("\nFehler bei C_FindObjects");
			if (ulMaxCount==0)
				printf("\nKeinen privaten Schl�ssel zum Zertifikat gefunden !");
			return -1;
	   } else {
	        printf("\nC_FindObjects erfolgreich, Key Handle: %lX", pObject);
	   }

	   returnCode = p11->C_FindObjectsFinal(hSession);

	   //War das OK?
	   if (returnCode != CKR_OK) {
	        printf("\nFehler bei C_FindObjectsFinal");
	   }else {
	        printf("\nC_FindObjectsFinal erfolgreich");
	   }

	   //////////////////////////////////////
	   //Encrypt beginnen
	   // pPublic = PublicKey
	   // pObject = PrivateKey
	   CK_MECHANISM mechanism;
	   CK_BYTE data[1024], data2[1024];
	   CK_ULONG lenData=85;
	   CK_ULONG lenData2=256;
	   mechanism.mechanism = CKM_RSA_PKCS;
	   mechanism.pParameter = NULL_PTR;
	   mechanism.ulParameterLen = 0;

	   for(int i=0; i<128; i++) data[i] = i % 256;

	   // EncryptionVorgang initialisieren
	   returnCode = p11->C_EncryptInit(hSession, &mechanism, pPublic);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei initialisieren");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

			return -1;
		 }
	   printf("\ninitialisieren erfolgreich");

	   returnCode = p11->C_EncryptUpdate(hSession, data, lenData,data2,&lenData2);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei C_EncryptUpdate");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

			return -1;
		 }
	   printf("\nC_EncryptUpdate erfolgreich !\n");

	   returnCode = p11->C_EncryptFinal(hSession, NULL, &lenData);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei C_EncryptFinal");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

			return -1;
		 }

	   printf("\nC_EncryptFinal erfolgreich !\n Encrypted Data:\n");	   
	   printBytes(data2,lenData2);
	   printf("\n");

	   // Decryption initialisieren
	   returnCode = p11->C_DecryptInit(hSession, &mechanism, pObject);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei C_DecryptInit");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

			return -1;
		 }
	   printf("\nC_DecryptInit erfolgreich");

	   //Decrypt
	   lenData = 1024;
	   returnCode = p11->C_DecryptUpdate(hSession, data2,lenData2,data, &lenData);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei C_DecryptUpdate");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

			return -1;
		 }
	   printf("\nC_DecryptUpdate erfolgreich !\n");

	   returnCode = p11->C_DecryptFinal(hSession, NULL, &lenData2);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
			printf("\nFehler bei C_DecryptFinal");
			p11->C_CloseSession(hSession);
			free(pSlotId);
			p11->C_Finalize(NULL_PTR);

			return -1;
		 }

	   printf("\nC_DecryptFinal erfolgreich !\n Decrypted Data:\n");
	   printBytes(data,lenData);
	   printf("\n");


	   //Session schlie�en
	   returnCode = p11->C_CloseSession(hSession);

	   //War das OK?
	   if (returnCode != CKR_OK)
		 {
	        printf("\nFehler bei C_CloseSession");
	        free(pSlotId);
	        p11->C_Finalize(NULL_PTR);

	        return -1;
		 }

	   //Speicher freigeben
	   free(pSlotId);

	   //Sitzung beenden
	   returnCode = p11->C_Finalize(NULL_PTR);

	   if (returnCode != CKR_OK)
	     {
	        printf("\nFehler bei C_Finalize");

	        return -1;
	     }

	   //Warten
	   printf("\nTest erfolgreich !");
	   return 0;
	
}

int pkInfo(CK_FUNCTION_LIST_PTR p11, void* hDll) {
	
	CK_SESSION_HANDLE hSession;
	
	CK_RV returnCode;

	returnCode = p11->C_Initialize(0);
	if (returnCode != CKR_OK)
		printf("\n!!!Error C_Initialize!!!");
	else
		printf("\n!!!C_Initialize OK!!!");

	CK_SLOT_ID pSlotId[10];
	CK_ULONG ulSlotCount = 10;
	returnCode = p11->C_GetSlotList(CK_FALSE,pSlotId,&ulSlotCount);
	if (returnCode != CKR_OK)
		printf("\n!!!Error C_Initialize!!!");
	else
		printf("\n!!!C_Initialize OK!!!");

	   //Daten der Slots anzeigen
	   for(unsigned long i=0; i<ulSlotCount; i++) {

		   CK_SLOT_INFO slotInfo;
		   returnCode = p11->C_GetSlotInfo(pSlotId[i], &slotInfo);
		   if (returnCode != CKR_OK) {
				printf("\nFehler bei C_GetSlotInfo");
				return -1;
		   }

		   //Infos zum Token holen
		   CK_TOKEN_INFO tokenInfo;
		   returnCode = p11->C_GetTokenInfo(pSlotId[i], &tokenInfo);
		   if (returnCode != CKR_OK) {
				printf("\nFehler bei C_GetTokenInfo");
				return -1;
		   }

		   //Daten ausgeben
		   printf("\n\nSlotId                     : %lx", pSlotId[i]);
		   printf("\nBeschreibung des Slots       : ");
		   for (unsigned int k=0; k<64; k++) printf("%c", slotInfo.slotDescription[k]);
		   printf("\nHersteller des Slots         : ");
		   for(unsigned int k=0; k<32; k++) printf("%c", slotInfo.manufacturerID[k]);
		   printf("\nVersion der Hardware         : %d.%d", slotInfo.hardwareVersion.major, slotInfo.hardwareVersion.minor);
		   printf("\nVersion der Firmware         : %d.%d", slotInfo.firmwareVersion.major, slotInfo.firmwareVersion.minor);
		   printf("\nBezeichnung des Tokens       : ");
		   for (unsigned int k=0; k<32; k++) printf("%c", tokenInfo.label[k]);
		   printf("\nHersteller des Tokens        : ");
		   for (unsigned int k=0; k<32; k++) printf("%c", tokenInfo.manufacturerID[k]);
		   printf("\nModell des Tokens            : ");
		   for (unsigned int k=0; k<16; k++) printf("%c", tokenInfo.model[k]);
		   printf("\nSeriennummer des Tokens      : ");
		   for (unsigned int k=0; k<16; k++) printf("%c", tokenInfo.serialNumber[k]);
		   printf("\nMax. Sessions                : %ld", tokenInfo.ulMaxSessionCount);
		   printf("\nVorhandene Sessions          : %ld", tokenInfo.ulSessionCount);
		   printf("\nMax. Sessions (RW)           : %ld", tokenInfo.ulMaxRwSessionCount);
		   printf("\nVorhandene Sessions (RW)     : %ld", tokenInfo.ulRwSessionCount);
		   printf("\nMax. Pin-Laenge              : %ld", tokenInfo.ulMaxPinLen);
		   printf("\nMin. Pin-Laenge              : %ld", tokenInfo.ulMinPinLen);
		   printf("\nOeffentlicher Speicher       : %ld", tokenInfo.ulTotalPublicMemory);
		   printf("\nFreier oeffentlicher Speicher: %ld", tokenInfo.ulFreePublicMemory);
		   printf("\nPrivater Speicher            : %ld", tokenInfo.ulTotalPrivateMemory);
		   printf("\nFreier privater Speicher     : %ld", tokenInfo.ulFreePrivateMemory);
		   printf("\nVersion der Token-Hardware   : %d.%d", tokenInfo.hardwareVersion.major, slotInfo.hardwareVersion.minor);
		   printf("\nVersion der Token-Firmware   : %d.%d", tokenInfo.firmwareVersion.major, slotInfo.firmwareVersion.minor);

		   //Session �ffnen
		   returnCode = p11->C_OpenSession(pSlotId[i], CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession);
		   if (returnCode != CKR_OK) {
				printf("\nFehler bei C_OpenSession");
				return -1;
		   }

		   if (true) {
				//Einloggen ohne PinPad
				returnCode = p11->C_Login(hSession, CKU_USER, ucPWDUser, dwLenPWDUser);
				//War das OK?
				if (returnCode != CKR_OK) {
					printf("\nFehler bei C_Login");
					return -1;
				} else {
					printf("\nC_Login erfolgreich, User: CKU_USER, PIN: '123456', Session: %lX",hSession);
				}
		   }

		   //SessionInformation anzeigen
		   CK_SESSION_INFO sessionInfo;
		   returnCode = p11->C_GetSessionInfo(hSession, &sessionInfo);
		   if (returnCode != CKR_OK) {
				printf("\nFehler bei C_GetSessionInfo");
				return -1;
		   }

		   printf("\nSlotId (SessionInfo)      : %lx", sessionInfo.slotID);
		   printf("\nFlags (SessionInfo)       : %lx", sessionInfo.flags);
		   printf("\nStatus (SessionInfo)      : %lx", sessionInfo.state);
		   printf("\nGer�tefehler (SessionInfo): %lx", sessionInfo.ulDeviceError);

		   //Objekte des Tokens ausgeben
		   returnCode = p11->C_FindObjectsInit(hSession, NULL_PTR, 0);
		   if (returnCode != CKR_OK) {
				printf("\nFehler bei C_FindObjectsInit");
				return -1;
		   }

		   //Suchen
		   CK_OBJECT_HANDLE objectList[20];
		   CK_ULONG objectsFound=0;

		   returnCode = p11->C_FindObjects(hSession, objectList, 20, &objectsFound);
		   if (returnCode != CKR_OK) {
				printf("\nFehler bei C_FindObjects");
				return -1;
		   }

		   returnCode = p11->C_FindObjectsFinal(hSession);
		   if (returnCode != CKR_OK) {
				printf("\nFehler bei C_FindObjectsFinal");
				return -1;
		   }

		   //Jetzt die Daten aller Objekte anzeigen
		   printf("\n\nObjekte des Tokens: ");

		   for(unsigned int i=0; i<objectsFound; i++) {

				 CK_ATTRIBUTE attr[20];
				 char objectLabel[100];

				 //Nur den Label holen
				 attr[0].type = CKA_LABEL;
				 attr[0].pValue = objectLabel;
				 attr[0].ulValueLen = sizeof(objectLabel);

				 //Attribute holen
				 returnCode = p11->C_GetAttributeValue(hSession, objectList[i], attr, 1);
				 if (returnCode != CKR_OK) {
					 printf("\nFehler bei C_GetAttributeValue");
					return -1;
				 }
				 //Ausgeben
				 printf("\n");
				 for(unsigned int j=0; j<attr[0].ulValueLen; j++) printf("%c", objectLabel[j]);
		   }

		   //Session beenden
		   returnCode = p11->C_CloseSession(hSession);

		   if (returnCode != CKR_OK) {
				printf("\nFehler bei C_CloseSession");
				return -1;
		   }
	   }

	  return 0;
}

int main(int argc, char* argv[]) {

	// 1. Argument = Test
	// 2. Argument = Library
	#ifndef _WIN32
	std::string sLibPath = "/home/malte/workspace/pkcs11tcos3/Debug/libpkcs11tcos3.so";
	#else
	std::string sLibPath = "P11TCOS3NetKey.dll";
	#endif

	if (argc > 1) {

		if (argc > 2) {
			sLibPath = argv[2];
		}

		char* pcTest = argv[1];
		if ( (pcTest[0] < 0x30) || (pcTest[0] > 0x34) ) {
			return -1;
		}
		
		void* hDll;
		CK_FUNCTION_LIST_PTR p11 = getP11(&hDll,sLibPath);
		int iReturn;

		// PKCS11 Modul Informationen ausgeben
		if (pcTest[0] == 0x30)
			iReturn = pkInfo(p11,hDll);
	
		// RSA Encrypt/Decrypt EXTENDED APDU LENGTH NOT SUPPORTED
		if (pcTest[0] == 0x31)
			iReturn = pkEncrypt(p11,hDll);
	
		// Login with PinPad not supported
		if (pcTest[0] == 0x32)
			iReturn = pkSign(p11,hDll,true);
	
		// Sign after standard login
		if (pcTest[0] == 0x33)
			iReturn = pkSign(p11,hDll,false);

		// SignRecover after standard login
		if (pcTest[0] == 0x34)
			iReturn = pkSignRecover(p11,hDll);
		
		freeDll(hDll);	
		return iReturn;

	} else {
		printf("\n 1. Argumet: Test");
		printf("\n\t 0 : Modulinfo-Test");
		printf("\n\t 1 : Encryption-Test");
		printf("\n\t 2 : Signatur/PinPad-Test");
		printf("\n\t 3 : Signatur-Test");
		printf("\n\t 4 : Signatur/Recover-Test");
		printf("\n\n 2. Argumet: PKCS#11 Library");
		printf("\n\t Default : %s",sLibPath.c_str());
		return 0;
	}

}
