//---------------------------------------------------------------------------

#include <windows.h>
#include <stdio.h>
#include <conio.h>

//Includes fr PKCS#11
#include "cryptoki.h"

#pragma hdrstop

//---------------------------------------------------------------------------

#pragma argsused
void printBytes(CK_BYTE_PTR pBytes, DWORD dLenBytes) {

	for (DWORD d = 0; d < dLenBytes; d++) {
		printf("0x%X ",(pBytes[d]));
	}
}
int main(int argc, char* argv[])
{

   unsigned char ucPWDUser[] = { 0x31, 0x32, 0x33, 0x34, 0x35, 0x36 };
   DWORD dwLenPWDUser = 6;

   CK_FUNCTION_LIST_PTR p11;
   CK_RV returnCode;
   CK_C_GetFunctionList C_GetFunctionListDyn;

   if (argc != 2) {
		printf("Bitte LibName als Argument angeben");
		exit(0);
   }
   //DLL laden
   HINSTANCE hDll = LoadLibrary(argv[1]);
   if (hDll == NULL) {
		printf("Fehler beim Laden der Bibliothek");
		exit(0);
   }

   //Funktionen holen
   C_GetFunctionListDyn = (CK_C_GetFunctionList) GetProcAddress(hDll, "C_GetFunctionList");

   //Liste der Funktionen holen
   returnCode = C_GetFunctionListDyn(&p11);
   if (returnCode != CKR_OK) {
		FreeLibrary(hDll);
		hDll = NULL;
		p11 = NULL;
		printf("\nFehler bei C_GetFunctionList (dyn)");
		exit(0);
   }

   //PKCS#11 initialisieren
   returnCode = p11->C_Initialize(NULL_PTR);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Initialize");
		exit(0);
   } else {
        printf("\nC_Initialize erfolgreich");
   }

   //Anzahl Slots holen
   CK_ULONG ulSlotCount;
   returnCode = p11->C_GetSlotList(CK_TRUE, NULL_PTR, &ulSlotCount);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetSlotList");
        exit(0);
   } else {
        printf("\nC_GetSlotList erfolgreich");
   }

   //Sind Slots da?
   if (ulSlotCount == 0) {
        printf("\nKeine Slots gefunden");
        exit(0);
   } else {
        printf("\n %d Slots gefunden !",ulSlotCount);
   }

   //Slots holen
   CK_SLOT_ID_PTR pSlotId = (CK_SLOT_ID_PTR) malloc(sizeof(CK_SLOT_ID) * ulSlotCount);
   returnCode = p11->C_GetSlotList(CK_TRUE, pSlotId, &ulSlotCount);
   if (returnCode != CKR_OK) {
        printf("\nFehler bei C_GetSlotList");
        exit(0);
   } else {
        printf("\nC_GetSlotList erfolgreich, verwende SlotID : %X",pSlotId[0]);
   }

   //1. Slot verwenden !!
   //Session ffnen
   CK_SESSION_HANDLE hSession;
   returnCode = p11->C_OpenSession(pSlotId[0], CKF_SERIAL_SESSION, "Test-InitPin", NULL_PTR, &hSession);
   //War das OK?
   if (returnCode != CKR_OK) {
        printf("\nFehler bei C_OpenSession");
        exit(0);
   } else {
		printf("\nC_OpenSession erfolgreich, Session Handle : %X",hSession);
   }

   //Einloggen ohne PinPad
   returnCode = p11->C_Login(hSession, CKU_USER, ucPWDUser, dwLenPWDUser);
   //War das OK?
   if (returnCode != CKR_OK) {
        printf("\nFehler bei C_Login");
        exit(0);
   } else {
		printf("\nC_Login erfolgreich, User: CKU_USER, PIN: '123456', Session: %X",hSession);
   }

   //////////////////////////////////////
   // Zertifikat suchen
   // Template fr X.509 Zertifikat
   CK_OBJECT_CLASS cClass   = CKO_PUBLIC_KEY;
   CK_BBOOL cbEncrypt = CK_TRUE;
   CK_ATTRIBUTE certTempl[]    = {
		{
			CKA_CLASS,
			&cClass,
			sizeof(cClass)
		},
		{
			CKA_ENCRYPT,
			 &cbEncrypt,
			sizeof(cbEncrypt)
		}
   };

   returnCode = p11->C_FindObjectsInit(hSession,certTempl,2);
   if (returnCode != CKR_OK) {
        printf("\nFehler bei C_FindObjectsInit");
        exit(0);
   } else {
		printf("\nC_FindObjectsInit erfolgreich, CKA_CLASS: CKO_PUBLIC_KEY, CKA_ENCRYPT: TRUE");
   }
   CK_OBJECT_HANDLE pPublic;
   CK_ULONG ulMaxCount;

   // Erstes gefundene Zertifikat verwenden in pCertificateHandle
   returnCode = p11->C_FindObjects(hSession,&pPublic,1,&ulMaxCount);
   if ( (returnCode != CKR_OK) || (ulMaxCount==0)) {
        printf("\nFehler bei C_FindObjects");
		exit(0);
   } else {
		printf("\nC_FindObjects erfolgreich, Public Key Handle: %X", pPublic);
   }

   // Suche abschliessen
   returnCode = p11->C_FindObjectsFinal(hSession);
   if (returnCode != CKR_OK) {
        printf("\nFehler bei C_FindObjectsFinal");
        exit(0);
   } else {
        printf("\nC_FindObjectsFinal erfolgreich");
   }

   // Zertifikatshandle nun in pCertificateHandle !!
   // Anhand dessen CKA_ID finden
   CK_BYTE pbCKA_ID[1024];
   memset(pbCKA_ID,0x00,1024);
   CK_ATTRIBUTE paKeyCKA_ID[]   = { { CKA_ID, NULL, 0 } };
   p11->C_GetAttributeValue(hSession, pPublic, paKeyCKA_ID, 1);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetAttributeValue");
		exit(0);
   } else {
		printf("\nC_GetAttributeValue erfolgreich, CKA_ID Lnge: %d", paKeyCKA_ID[0].ulValueLen);
   }
   DWORD dLenID = paKeyCKA_ID[0].ulValueLen;
   CK_ATTRIBUTE pKeyCKA_ID[] = { { CKA_ID, pbCKA_ID, dLenID } };
   p11->C_GetAttributeValue(hSession, pPublic, pKeyCKA_ID, 1);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetAttributeValue");
		exit(0);
   } else {
		printf("\nC_GetAttributeValue erfolgreich, CKA_ID :");
		printBytes(pbCKA_ID,dLenID);
		printf("\n");
   }

   //////////////////////////////////////
   // PrivateKey anhand der CKA_ID des Zertifikats finden
   cClass = CKO_PRIVATE_KEY;
   CK_ATTRIBUTE keyTempl[]    = {
		{
			CKA_CLASS,
			(unsigned char *) &cClass,
			sizeof(cClass)
		},
		{
			CKA_ID,
		   (unsigned char *) &pbCKA_ID,
			dLenID
		}
   };

   returnCode = p11->C_FindObjectsInit(hSession,keyTempl,2);
   if (returnCode != CKR_OK) {
        printf("\nFehler bei C_FindObjectsInit");
        exit(0);
   } else {
		printf("\nC_FindObjectsInit erfolgreich, CKA_CLASS: CKO_PRIVATE_KEY CKA_ID: ");
		printBytes(pbCKA_ID,dLenID);
		printf("\n");
   }

   CK_OBJECT_HANDLE pObject;
   ulMaxCount;
   returnCode = p11->C_FindObjects(hSession,&pObject,1,&ulMaxCount);
   if ( (returnCode != CKR_OK) || (ulMaxCount==0))  {
		printf("\nFehler bei C_FindObjects");
		if (ulMaxCount==0)
			printf("\nKeinen privaten Schlssel zum Zertifikat gefunden !");
		exit(0);
   } else {
        printf("\nC_FindObjects erfolgreich, Key Handle: %X", pObject);
   }

   returnCode = p11->C_FindObjectsFinal(hSession);

   //War das OK?
   if (returnCode != CKR_OK) {
        printf("\nFehler bei C_FindObjectsFinal");
   }else {
        printf("\nC_FindObjectsFinal erfolgreich");
   }

   //////////////////////////////////////
   //Encrypt beginnen
   // pPublic = PublicKey
   // pObject = PrivateKey
   CK_MECHANISM mechanism;
   CK_OBJECT_HANDLE hKey = pObject;
   CK_BYTE data[1024], data2[1024], sig[256];
   CK_ULONG lenData=85;
   CK_ULONG lenData2=256;
   mechanism.mechanism = CKM_RSA_PKCS;
   mechanism.pParameter = NULL_PTR;
   mechanism.ulParameterLen = 0;

   for(int i=0; i<128; i++) data[i] = i % 256;

   // EncryptionVorgang initialisieren
   returnCode = p11->C_EncryptInit(hSession, &mechanism, pPublic);

   //War das OK?
   if (returnCode != CKR_OK)
	 {
		printf("\nFehler bei initialisieren");
		p11->C_CloseSession(hSession);
		free(pSlotId);
		p11->C_Finalize(NULL_PTR);

		exit(0);
	 }
   printf("\initialisieren erfolgreich");

   returnCode = p11->C_EncryptUpdate(hSession, data, lenData,data2,&lenData2);

   //War das OK?
   if (returnCode != CKR_OK)
	 {
		printf("\nFehler bei C_EncryptUpdate");
		p11->C_CloseSession(hSession);
		free(pSlotId);
		p11->C_Finalize(NULL_PTR);

		exit(0);
	 }
   printf("\C_EncryptUpdate erfolgreich !\n");

   returnCode = p11->C_EncryptFinal(hSession, NULL, &lenData);

   //War das OK?
   if (returnCode != CKR_OK)
	 {
		printf("\nFehler bei C_EncryptFinal");
		p11->C_CloseSession(hSession);
		free(pSlotId);
		p11->C_Finalize(NULL_PTR);

		exit(0);
	 }

   printf("\C_EncryptFinal erfolgreich !\n Encrypted Data:\n");
   printBytes(data2,lenData2);
   printf("\n");

   // Decryption initialisieren
   returnCode = p11->C_DecryptInit(hSession, &mechanism, pObject);

   //War das OK?
   if (returnCode != CKR_OK)
	 {
		printf("\nFehler bei C_DecryptInit");
		p11->C_CloseSession(hSession);
		free(pSlotId);
		p11->C_Finalize(NULL_PTR);

		exit(0);
	 }
   printf("\C_DecryptInit erfolgreich");

   //Decrypt
   lenData = 1024;
   returnCode = p11->C_DecryptUpdate(hSession, data2,lenData2,data, &lenData);

   //War das OK?
   if (returnCode != CKR_OK)
	 {
		printf("\nFehler bei C_DecryptUpdate");
		p11->C_CloseSession(hSession);
		free(pSlotId);
		p11->C_Finalize(NULL_PTR);

		exit(0);
	 }
   printf("\C_DecryptUpdate erfolgreich !\n");

   returnCode = p11->C_DecryptFinal(hSession, NULL, &lenData2);

   //War das OK?
   if (returnCode != CKR_OK)
	 {
		printf("\nFehler bei C_DecryptFinal");
		p11->C_CloseSession(hSession);
		free(pSlotId);
		p11->C_Finalize(NULL_PTR);

		exit(0);
	 }

   printf("\C_DecryptFinal erfolgreich !\n Decrypted Data:\n");
   printBytes(data,lenData);
   printf("\n");


   //Session schlieen
   returnCode = p11->C_CloseSession(hSession);

   //War das OK?
   if (returnCode != CKR_OK)
	 {
        printf("\nFehler bei C_CloseSession");
        free(pSlotId);
        p11->C_Finalize(NULL_PTR);

        exit(0);
	 }

   //Speicher freigeben
   free(pSlotId);

   //Sitzung beenden
   returnCode = p11->C_Finalize(NULL_PTR);

   if (returnCode != CKR_OK)
     {
        printf("\nFehler bei C_Finalize");

        exit(0);
     }

   //Warten
   printf("\nTest erfolgreich !");
   getch();

}
//---------------------------------------------------------------------------
